"""Low-level SLURM API to submit, cancel and monitor jobs"""

import os
import time
import json
import requests
import logging
from pprint import pformat
from dataclasses import dataclass
from contextlib import contextmanager
from typing import Iterator, List, Optional, Tuple, Union, Generator

from .. import defaults
from .. import os_utils
from .. import utils
from ..errors import RemoteHttpError
from ..errors import raise_chained_errors
from ..log_utils import log_file_monitor


logger = logging.getLogger(__name__)


class SlurmBaseRestClient:
    """Low-level SLURM API to submit, cancel and monitor jobs.
    This class does not contain any job-related state."""

    VERSION = "0.0.37"
    FINISHED_STATES = "FAILED", "COMPLETED", "CANCELLED", "TIMEOUT"

    def __init__(
        self,
        url: str = "",
        user_name: str = "",
        token: str = "",
        parameters: Optional[dict] = None,
        log_directory: Optional[str] = None,
        std_split: Optional[bool] = False,
        request_options: Optional[dict] = None,
    ):
        """
        :param url: SLURM REST API
        :param user_name: User name on SLURM
        :param token: SLURM access token for the user
        :param parameters: SLURM job parameters
        :param log_directory: SLURM log directory
        :param std_split: Split standard output and standard error
        :param request_options: GET, POST and DELETE options
        """
        if not url:
            raise ValueError("Requires the SLURM REST url")
        if not token:
            raise ValueError("Requires a SLURM user token")
        if not user_name:
            raise ValueError("Requires a SLURM user name")
        self._url = url
        self._user_name = user_name.rstrip()
        self._token = token.replace("SLURM_JWT=", "").rstrip()
        self._parameters = parameters
        if log_directory:
            log_directory = log_directory.format(user_name=user_name)
        self._log_directory = log_directory
        self._std_split = std_split
        self._request_options = request_options

    def get_headers(self) -> dict:
        return {"X-SLURM-USER-NAME": self._user_name, "X-SLURM-USER-TOKEN": self._token}

    def get(
        self,
        path: str,
        request_options: Optional[dict] = None,
        error_msg: Optional[str] = None,
        raise_on_error: bool = True,
    ) -> dict:
        request_options = self._parse_request_options(request_options)
        response = requests.get(self._url + path, **request_options)
        return self._handle_response(
            response, raise_on_error=raise_on_error, error_msg=error_msg
        )

    def delete(
        self,
        path: str,
        request_options: Optional[dict] = None,
        error_msg: Optional[str] = None,
        raise_on_error: bool = True,
    ) -> dict:
        request_options = self._parse_request_options(request_options)
        response = requests.delete(self._url + path, **request_options)
        return self._handle_response(
            response, raise_on_error=raise_on_error, error_msg=error_msg
        )

    def post(
        self,
        path: str,
        json: Optional[dict] = None,
        request_options: Optional[dict] = None,
        raise_on_error: bool = True,
        error_msg: Optional[str] = None,
    ) -> dict:
        request_options = self._parse_request_options(request_options)
        if json:
            request_options["json"] = json
            logger.debug("POST %s\n%s", path, pformat(json))
        else:
            logger.debug("POST %s", path)
        response = requests.post(self._url + path, **request_options)
        return self._handle_response(
            response, raise_on_error=raise_on_error, error_msg=error_msg
        )

    def _parse_request_options(self, request_options: Optional[dict]) -> dict:
        merged = utils.merge_mappings(self._request_options, request_options)
        if merged.get("headers") is None:
            merged["headers"] = self.get_headers()
        return merged

    def server_has_api(self, request_options: Optional[dict] = None) -> bool:
        response = self.get("/openapi", request_options=request_options)
        return self.VERSION == response["info"]["version"]

    def submit_job(
        self,
        script: str,
        parameters: Optional[dict] = None,
        metadata: Optional[Union[str, dict]] = None,
        request_options: Optional[dict] = None,
    ) -> int:
        """Returns the SLURM job ID"""
        # https://slurm.schedmd.com/rest_api.html#slurmctldSubmitJob
        job = self._create_job_properties(parameters=parameters, metadata=metadata)
        job_recipe = {"script": script, "job": job}
        job_name = job["name"]

        response = self.post(
            f"/slurm/v{self.VERSION}/job/submit",
            json=job_recipe,
            request_options=request_options,
            error_msg=f"failed to submit SLURM job '{job_name}'",
        )

        job_id = response["job_id"]
        user_msg = response.get("job_submit_user_msg", "")
        if user_msg:
            logger.info(
                "SLURM job '%s' submitted (Job ID: %s): %s", job_name, job_id, user_msg
            )
        else:
            logger.info("SLURM job '%s' submitted (Job ID: %s)", job_name, job_id)
        return job_id

    def cancel_job(self, job_id: int, request_options: Optional[dict] = None) -> None:
        # https://slurm.schedmd.com/rest_api.html#slurmctldCancelJob
        self.delete(
            f"/slurm/v{self.VERSION}/job/{job_id}", request_options=request_options
        )

    def clean_job_artifacts(
        self,
        job_id: int,
        properties: Optional[dict] = None,
        request_options: Optional[dict] = None,
    ) -> None:
        if properties is None:
            properties = self.get_job_properties(
                job_id, raise_on_error=False, request_options=request_options
            )
        if properties is None:
            return
        self._cleanup_job_log_artifact(job_id, properties.get("standard_output"))
        self._cleanup_job_log_artifact(job_id, properties.get("standard_error"))

    @contextmanager
    def clean_job_artifacts_context(
        self,
        job_id: int,
        properties: Optional[dict] = None,
        request_options: Optional[dict] = None,
    ) -> Generator[None, None, None]:
        try:
            yield
        finally:
            self.clean_job_artifacts(
                job_id,
                properties=properties,
                request_options=request_options,
            )

    def get_job_properties(
        self, job_id: int, raise_on_error=True, request_options: Optional[dict] = None
    ) -> Optional[dict]:
        # https://slurm.schedmd.com/rest_api.html#slurmctldGetJob
        response = self.get(
            f"/slurm/v{self.VERSION}/job/{job_id}",
            request_options=request_options,
            raise_on_error=raise_on_error,
            error_msg=f"Failed to get properties of SLURM Job ID {job_id}",
        )
        jobs = response.get("jobs", list())
        if len(jobs) != 1:
            return None
        return jobs[0]

    def get_all_job_properties(
        self,
        raise_on_error=True,
        request_options: Optional[dict] = None,
        filter=None,
        update_time=None,
    ) -> List[dict]:
        # https://slurm.schedmd.com/rest_api.html#slurmctldGetJob
        path = f"/slurm/v{self.VERSION}/jobs"
        if update_time:
            path = f"{path}?update_time={update_time.timestamp()}"
        response = self.get(
            path,
            request_options=request_options,
            raise_on_error=raise_on_error,
            error_msg="failed to get SLURM job properties",
        )
        jobs = response.get("jobs", list())
        return list(self._filter_job_properties(jobs, filter))

    def get_stdout_stderr(
        self, job_id: int, raise_on_error=True, request_options: Optional[dict] = None
    ) -> Tuple["_StdInfo", Optional["_StdInfo"]]:
        properties = self.get_job_properties(
            job_id, raise_on_error=raise_on_error, request_options=request_options
        )
        if properties is None:
            properties = dict()
        ofilename = properties.get("standard_output")
        ofilename, ocontent = self._read_log(job_id, ofilename)
        efilename = properties.get("standard_error")
        efilename, econtent = self._read_log(job_id, efilename)
        if efilename is None:
            std = _StdInfo(name="STDOUT/STDERR", filename=ofilename, content=ocontent)
            return std, None
        else:
            stdout = _StdInfo(name="STDOUT", filename=ofilename, content=ocontent)
            stderr = _StdInfo(name="STDERR", filename=efilename, content=econtent)
            return stdout, stderr

    def get_status(self, job_id: int, request_options=None) -> str:
        # https://curc.readthedocs.io/en/latest/running-jobs/squeue-status-codes.html
        properties = self.get_job_properties(job_id, request_options=request_options)
        return properties.get("job_state", "UNKNOWN")

    def get_full_status(self, job_id: int, request_options=None) -> dict:
        properties = self.get_job_properties(job_id, request_options=request_options)
        return {
            "status": properties.get("job_state", "UNKNOWN"),
            "description": properties.get("state_description", ""),
            "reason": properties.get("state_reason", ""),
            "exit_code": self._parse_exit_code(properties.get("exit_code", "")),
        }

    def is_finished(self, job_id: int) -> bool:
        return self.get_status(job_id) in self.FINISHED_STATES

    def wait_finished(
        self,
        job_id: int,
        progress: bool = False,
        timeout: Optional[float] = None,
        period: float = 0.5,
    ) -> str:
        job_state = None
        t0 = time.time()
        while job_state not in self.FINISHED_STATES:
            job_state = self.get_status(job_id)
            time.sleep(period)
            if progress:
                print(".", end="", flush=True)
            if timeout is not None and (time.time() - t0) > timeout:
                raise TimeoutError
        status = self.get_full_status(job_id)
        logger.info("Job '%s' finished: %s", job_id, pformat(status))
        return job_state

    def print_stdout_stderr(self, job_id: int):
        stdout, stderr = self.get_stdout_stderr(job_id, raise_on_error=False)
        print()
        print("".join(stdout.lines()))
        if stderr is not None:
            print("".join(stderr.lines()))

    def log_stdout_stderr(
        self, job_id: int, logger=None, level=logging.INFO, **log_options
    ) -> None:
        if logger is None:
            logger = logger
        stdout, stderr = self.get_stdout_stderr(job_id, raise_on_error=False)
        stdout = "".join(stdout.lines())
        if stderr is None:
            logger.log(level, "\n%s", stdout, **log_options)
        else:
            stderr = "".join(stderr.lines())
            logger.log(level, "\n%s\n%s", stdout, stderr, **log_options)

    @contextmanager
    def redirect_stdout_stderr(
        self, job_id: int, request_options: Optional[dict] = None
    ) -> Generator[None, None, None]:
        """Redirect logs files to the local root logger within this context."""
        properties = self.get_job_properties(
            job_id, raise_on_error=False, request_options=request_options
        )
        if properties is None:
            properties = dict()
        ofilename = properties.get("standard_output")
        efilename = properties.get("standard_error")
        ofilename = self._parse_filename(job_id, ofilename)
        efilename = self._parse_filename(job_id, efilename)
        with log_file_monitor([ofilename, efilename]):
            yield

    def _handle_response(
        self, response, raise_on_error: bool = True, error_msg: Optional[str] = None
    ):
        try:
            response.raise_for_status()
        except Exception as e:
            if response.status_code in (400, 500):
                if not raise_on_error:
                    return dict()
                if not error_msg:
                    error_msg = str(e)
                if response.status_code == 500:
                    error_msg = f"{error_msg} (SLURM token expired?)"
                else:
                    error_msg = f"{error_msg} (Invalid SLURM parameter?)"
                raise RemoteHttpError(error_msg)
            raise
        data = response.json()
        if raise_on_error:
            self._raise_on_error(data.get("errors"), error_msg=error_msg)
        return data

    def _filter_job_properties(
        self, jobs: List[dict], filter: Optional[dict]
    ) -> Iterator[dict]:
        if filter is None:
            filter = dict()
        filter.setdefault("user_name", self._user_name)
        filter = {k: v for k, v in filter.items() if v is not None}
        for properties in jobs:
            if self._match_filter(properties, None, filter):
                yield properties

    def _match_filter(
        self, properties: dict, name: str, value: Optional[Union[str, dict]]
    ) -> bool:
        if name is None:
            return all(
                self._match_filter(properties, name, value)
                for name, value in value.items()
            )
        elif isinstance(value, str):
            return properties.get(name) == value
        else:
            properties = properties.get(name, dict())
            return all(self._match_filter(properties, k, v) for k, v in value.items())

    def _parse_exit_code(self, exit_code: Optional[int]) -> Optional[int]:
        if not exit_code:
            return
        return exit_code // 256

    def _read_log(
        self, job_id: int, filename: Optional[str]
    ) -> Tuple[str, Optional[List[str]]]:
        filename = self._parse_filename(job_id, filename)
        if filename is None:
            return None, None
        try:
            with open(filename) as f:
                return filename, list(f)
        except FileNotFoundError:
            return None, None

    def _cleanup_job_log_artifact(self, job_id: int, filename: Optional[str]) -> None:
        filename = self._parse_filename(job_id, filename)
        if filename is None or filename == "/dev/null":
            return None
        try:
            os.remove(filename)
        except (FileNotFoundError, PermissionError):
            return None

    def _parse_filename(self, job_id: int, filename: Optional[str]) -> Optional[str]:
        if not filename:
            return None
        return filename.replace("%j", str(job_id))

    def _raise_on_error(
        self, errors: Optional[List[dict]], error_msg: Optional[str] = None
    ):
        if not errors:
            return
        errors = [self._parse_error(error) for error in errors]
        if error_msg:
            errors.append(error_msg + " (token expired?)")
        raise_chained_errors(errors)

    def _parse_error(self, error: dict) -> str:
        description = error.get("error", "<no error description>")
        description = error.get("description", description)
        error_code = error.get("error_code", "<no error code>")
        error_code = error.get("error_number", description)
        return f"{description} (error code: {error_code})"

    def _create_job_properties(
        self,
        parameters: Optional[dict] = None,
        metadata: Optional[Union[str, dict]] = None,
    ) -> dict:
        # https://slurm.schedmd.com/rest_api.html#v0.0.37_job_properties
        parameters = utils.merge_mappings(self._parameters, parameters)
        if not parameters.get("environment"):
            # Cannot be empty
            parameters["environment"] = {"_DUMMY_VAR": "dummy_value"}

        job_name = self._ensure_job_name(parameters)

        parameters["standard_input"] = "/dev/null"
        if self._log_directory:
            os_utils.makedirs(self._log_directory)
            filetemplate = f"{self._log_directory}/{job_name}.%j"
            if self._std_split:
                parameters["standard_output"] = filetemplate + ".out"
                parameters["standard_error"] = filetemplate + ".err"
            else:
                parameters["standard_output"] = filetemplate + ".outerr"
        else:
            parameters["standard_output"] = "/dev/null"
            parameters["standard_error"] = "/dev/null"

        if metadata:
            if not isinstance(metadata, str):
                metadata = json.dumps(metadata)
            parameters["comment"] = metadata

        return parameters

    def _ensure_job_name(self, parameters: dict) -> str:
        job_name = parameters.get("name")
        if not job_name:
            job_name = defaults.JOB_NAME
            parameters["name"] = job_name
        return job_name


@dataclass
class _StdInfo:
    """Information about a SLURM job stdout/stderr log file"""

    name: str
    filename: Optional[str]
    content: Optional[List[str]]

    def lines(self) -> List[str]:
        title = f"{self.name}: {self.filename}\n"
        line = "-" * len(title) + "\n"
        if self.content is None:
            return [title, line, " <not found>"]
        elif not self.content:
            return [title, line, " <empty>"]
        else:
            return [title, line] + self.content



import os
import sys
import logging
import subprocess
from tempfile import NamedTemporaryFile
from typing import Optional
from unittest.mock import patch
from contextlib import ExitStack, contextmanager
from concurrent.futures import ThreadPoolExecutor

import pytest
from ..client import SlurmBaseRestClient
from ..client import SlurmScriptRestClient
from ..client import SlurmPyConnRestClient
from ..client.errors import RemoteHttpError

logger = logging.getLogger(__name__)

_MAX_JOBS = 50


@contextmanager
def mock_slurm_clients(tmpdir):
    last_job_id = 0
    jobs = dict()

    def _get(
        self,
        path: str,
        request_options=None,
        error_msg: Optional[str] = None,
        raise_on_error: bool = True,
    ):
        if path == "/openapi":
            return {"info": {"version": SlurmBaseRestClient.VERSION}}
        elif path.startswith("/slurm/v0.0.37/job/"):
            job_id = int(path.split("/")[-1])
            job_info = jobs.get(job_id)
            if job_info is None:
                if raise_on_error:
                    raise RemoteHttpError(error_msg)
                else:
                    return dict()
            logger.debug("SLURM mock backend: get job %s", job_id)
            return {"jobs": [job_info]}
        elif path == "/slurm/v0.0.37/jobs":
            logger.debug("SLURM mock backend: get jobs (# %d)", len(jobs))
            return {"jobs": list(jobs.values())}
        else:
            raise NotImplementedError(path)

    def _delete(
        self,
        path: str,
        request_options=None,
        error_msg: Optional[str] = None,
        raise_on_error: bool = True,
    ):
        if path.startswith("/slurm/v0.0.37/job/"):
            job_id = int(path.split("/")[-1])
            job_info = jobs.get(job_id)
            if job_info is None:
                if raise_on_error:
                    raise RemoteHttpError(error_msg)
            else:
                job_info["job_state"] = "CANCELLED"
                logger.debug("SLURM mock backend: cancel job %s", last_job_id)
        else:
            raise NotImplementedError(path)

    def _post(
        self,
        path: str,
        json: Optional[dict] = None,
        request_options=None,
        raise_on_error: bool = True,
        error_msg: Optional[str] = None,
    ):
        nonlocal last_job_id
        if path == "/slurm/v0.0.37/job/submit":
            last_job_id += 1
            job_info = {
                "job_id": last_job_id,
                "job_state": "PENDING",
                **json["job"],
                "user_name": self._user_name,
            }
            jobs[last_job_id] = job_info
            logger.debug("SLURM mock backend: recieved job %s", last_job_id)

            lines = json["script"].split("\n")
            shebang = lines[0]
            if "bash" in shebang:
                if sys.platform == "win32":
                    pytest.skip("bash script does not run on windows")
            elif "python" in shebang:
                pass
            else:
                assert False, f"Unknown script starting with '{shebang}'"

            pool.submit(_job_main, json, job_info, str(tmpdir))
            return job_info
        else:
            raise NotImplementedError(path)

    with ExitStack() as stack:
        for cls in (
            SlurmBaseRestClient,
            SlurmScriptRestClient,
            SlurmPyConnRestClient,
        ):
            ctx = ThreadPoolExecutor(max_workers=_MAX_JOBS)
            pool = stack.enter_context(ctx)
            ctx = patch.object(cls, "get", _get)
            stack.enter_context(ctx)
            ctx = patch.object(cls, "post", _post)
            stack.enter_context(ctx)
            ctx = patch.object(cls, "delete", _delete)
            stack.enter_context(ctx)
        yield


def _job_main(json: dict, job_info: dict, tmpdir):
    cmd = []
    lines = json["script"].split("\n")
    shebang = lines[0]
    if "bash" in shebang:
        suffix = ".sh"
    elif "python" in shebang:
        if sys.platform == "win32":
            lines.pop(0)
            cmd = [sys.executable]
        suffix = ".py"
    else:
        assert False, f"Unknown script starting with '{shebang}'"

    with NamedTemporaryFile("w", delete=False, dir=tmpdir, suffix=suffix) as script:
        script.write("\n".join(lines))
        filename = script.name
    os.chmod(filename, 0o755)
    cmd.append(filename)

    env = json["job"].get("environment", dict())
    env = {k: str(v) for k, v in env.items()}
    env = {**os.environ, **env}
    env["SLURM_JOB_ID"] = str(job_info["job_id"])

    standard_output = json["job"].get("standard_output")
    standard_error = json["job"].get("standard_error")

    if standard_output is None:
        stdout = None
    elif standard_output == "/dev/null":
        stdout = None
        standard_output = None
    else:
        stdout = subprocess.PIPE

    if standard_error is None:
        stderr = None
    elif standard_error == "/dev/null":
        stderr = None
        standard_error = None
    else:
        stderr = stdout or subprocess.PIPE

    with subprocess.Popen(
        cmd, stdout=stdout, stderr=stderr, env=env, cwd=os.getcwd()
    ) as proc:
        logger.debug("SLURM mock backend: job %s started", job_info["job_id"])
        job_info["job_state"] = "RUNNING"
        outs, errs = proc.communicate(timeout=15)

        if standard_output is not None:
            outfile = standard_output.replace("%j", str(job_info["job_id"]))
            with open(outfile, "wb") as f:
                f.write(outs)

        if standard_error is not None:
            errfile = standard_error.replace("%j", str(job_info["job_id"]))
            with open(errfile, "wb") as f:
                f.write(errs)

        if job_info["job_state"] != "CANCELLED":
            if proc.returncode:
                job_info["job_state"] = "FAILED"
                logger.debug("SLURM mock backend: job %s failed", job_info["job_id"])
            else:
                job_info["job_state"] = "COMPLETED"
                logger.debug("SLURM mock backend: job %s completed", job_info["job_id"])



import os


def chmod(path) -> None:
    try:
        original_umask = os.umask(0)
        return os.chmod(path, mode=0o777)
    finally:
        os.umask(original_umask)


def makedirs(dirname: str) -> None:
    try:
        original_umask = os.umask(0)
        return os.makedirs(dirname, mode=0o777, exist_ok=True)
    finally:
        os.umask(original_umask)



"""Monitor SLURM jobs and status"""

import sys
import time
import argparse
import datetime
import logging
from typing import List, Any
from tabulate import tabulate

from .cli import common as common_cli
from .cli import status as status_cli
from .cli import cancel as cancel_cli
from .cli import submit as submit_cli
from ..client import SlurmScriptRestClient


logger = logging.getLogger(__name__)


def main(argv=None):
    if argv is None:
        argv = sys.argv

    parser = argparse.ArgumentParser(
        description="SLURM Job Monitor", prog="pyslurmutils"
    )
    subparsers = parser.add_subparsers(help="Commands", dest="command")

    check = subparsers.add_parser("check", help="Check slurm connection")
    common_cli.add_parameters(check)

    status = subparsers.add_parser("status", help="Job status")
    common_cli.add_parameters(status)
    status_cli.add_parameters(status)

    status = subparsers.add_parser("cancel", help="Cancel SLURM jobs")
    common_cli.add_parameters(status)
    cancel_cli.add_parameters(status)

    status = subparsers.add_parser("submit", help="Submit SLURM jobs")
    common_cli.add_parameters(status)
    submit_cli.add_parameters(status)

    args = parser.parse_args(argv[1:])

    if args.command == "status":
        command_status(args)
    elif args.command == "check":
        command_check(args)
    elif args.command == "cancel":
        command_cancel(args)
    elif args.command == "submit":
        command_submit(args)
    else:
        parser.print_help()
    return 0


def command_status(args):
    common_cli.apply_parameters(args)
    status_cli.apply_parameters(args)

    client = SlurmScriptRestClient(
        args.url,
        args.user_name,
        args.token,
        log_directory=args.log_directory,
    )
    for _ in _monitor_loop(args.interval):
        _print_jobs(client, args.jobid, args.all)
        if args.jobid:
            client.print_stdout_stderr(args.jobid)


def command_check(args):
    common_cli.apply_parameters(args)

    client = SlurmScriptRestClient(
        args.url,
        args.user_name,
        args.token,
        log_directory=args.log_directory,
    )
    assert client.server_has_api(), "Wrong Rest API version"


def command_cancel(args):
    common_cli.apply_parameters(args)
    cancel_cli.apply_parameters(args)
    client = SlurmScriptRestClient(
        args.url,
        args.user_name,
        args.token,
        log_directory=args.log_directory,
    )
    job_ids = args.job_ids
    if not job_ids:
        job_ids = [prop["job_id"] for prop in client.get_all_job_properties()]
    for job_id in job_ids:
        client.cancel_job(job_id)


def command_submit(args):
    common_cli.apply_parameters(args)
    submit_cli.apply_parameters(args)
    client = SlurmScriptRestClient(
        args.url,
        args.user_name,
        args.token,
        parameters=args.parameters,
        log_directory=args.log_directory,
    )
    job_id = client.submit_script(args.script)
    print(f"SLURM job {job_id} started")
    try:
        client.wait_finished(job_id)
        client.print_stdout_stderr(job_id)
    finally:
        client.clean_job_artifacts(job_id)


def _monitor_loop(interval):
    try:
        if not interval:
            yield
            return
        while True:
            yield
            time.sleep(interval)
    except KeyboardInterrupt:
        pass


def _print_jobs(client, jobid, all_users):
    if jobid:
        jobs = [client.get_job_properties(jobid)]
    else:
        if all_users:
            filter = {"user_name": None}
        else:
            filter = None
        jobs = client.get_all_job_properties(filter=filter)

    columns = {
        "ID": (_passthrough, ("job_id",)),
        "Name": (_passthrough, ("name",)),
        "State": (_passthrough, ("job_state",)),
        "User": (_passthrough, ("user_name",)),
        "Limit": (_time_limit, ("time_limit",)),
        "Pending": (_pending_time, ("submit_time", "start_time")),
        "Runtime": (_running_time, ("start_time", "end_time", "job_state")),
        "Resources": (_resources, ("partition", "tres_req_str", "tres_alloc_str")),
    }
    rows = list()
    for info in jobs:
        rows.append(
            [
                parser(*[info[k] if isinstance(k, str) else k for k in parser_args])
                for parser, parser_args in columns.values()
            ]
        )
    if not rows:
        return
    titles = list(columns)
    print(_format_info(titles, rows))


def _passthrough(x: Any) -> str:
    return str(x)


def _time_limit(x: int) -> str:
    return str(datetime.timedelta(minutes=x))


def _resources(partition: str, requested: str, allocated: str) -> str:
    return f"{partition}: {requested or allocated}"


def _pending_time(submit_time: int, start_time: int) -> str:
    return _time_diff(submit_time, start_time)


def _running_time(start_time: int, end_time: int, state: str) -> str:
    if state == "PENDING":
        return "-"
    if state in ("RUNNING", "COMPLETING"):
        end_time = 0
    return _time_diff(start_time, end_time)


def _time_diff(t0: int, t1: int) -> str:
    duration = _get_datetime(t1) - _get_datetime(t0)
    if duration.total_seconds() < 0:
        return "-"
    return str(duration)


def _get_datetime(epoch: int) -> datetime.datetime:
    if epoch == 0:
        now = datetime.datetime.now()
        return now.replace(microsecond=0)
    else:
        return datetime.datetime.fromtimestamp(epoch)


def _format_info(titles: List[str], rows: List[List[str]]) -> str:
    maxcolwidths = [None] * len(titles)
    maxcolwidths[1] = 40
    return tabulate(rows, headers=titles, maxcolwidths=maxcolwidths)


if __name__ == "__main__":
    sys.exit(main())



"""SLURM API to submit, cancel and monitor scripts that start a python process
to establish a connection over which python functions can be executed."""

import logging
from typing import Optional, Union

from .script import SlurmScriptRestClient
from ..job_io.local import RemoteWorkerProxy

from .. import defaults

logger = logging.getLogger(__name__)


class SlurmPyConnRestClient(SlurmScriptRestClient):
    """SLURM API to submit, cancel and monitor scripts that start a python process
    to establish a connection over which python functions can be executed.
    This class does not contain any job-related state."""

    def __init__(
        self,
        url: str = "",
        user_name: str = "",
        token: str = "",
        parameters: Optional[dict] = None,
        log_directory: Optional[str] = None,
        std_split: Optional[bool] = False,
        request_options: Optional[dict] = None,
        pre_script: Optional[str] = None,
        post_script: Optional[str] = None,
        python_cmd: Optional[str] = None,
    ):
        """
        :param url: SLURM REST API
        :param user_name: User name on SLURM
        :param token: SLURM access token for the user
        :param parameters: SLURM job parameters
        :param log_directory: SLURM log directory
        :param std_split: Split standard output and standard error
        :param request_options: GET, POST and DELETE options
        :param pre_script: Shell script to execute at the start of a job
        :param post_script: Shell script to execute at the end of a job
        :param python_cmd: Python command
        """
        self.pre_script = pre_script
        self.post_script = post_script
        self.python_cmd = python_cmd
        super().__init__(
            url=url,
            user_name=user_name,
            token=token,
            parameters=parameters,
            log_directory=log_directory,
            std_split=std_split,
            request_options=request_options,
        )

    def submit_script(
        self,
        worker_proxy: RemoteWorkerProxy,
        pre_script: Optional[str] = None,
        post_script: Optional[str] = None,
        python_cmd: Optional[str] = None,
        parameters: Optional[dict] = None,
        metadata: Optional[Union[str, dict]] = None,
        request_options: Optional[dict] = None,
    ) -> int:
        """Submit a script that will establish a connection initialized in the current process."""
        if parameters is None:
            parameters = dict()
        parameters.setdefault("environment", dict()).update(
            worker_proxy.remote_environment
        )

        if not metadata:
            metadata = dict()
        metadata.update(worker_proxy.metadata)

        script = self._make_executable(
            worker_proxy.remote_script(),
            pre_script=pre_script,
            post_script=post_script,
            python_cmd=python_cmd,
        )

        return super().submit_script(
            script=script,
            parameters=parameters,
            metadata=metadata,
            request_options=request_options,
        )

    def _make_executable(
        self,
        python_script: str,
        pre_script: Optional[str] = None,
        post_script: Optional[str] = None,
        python_cmd: Optional[str] = None,
    ) -> str:
        """Make a python script executable."""
        if not pre_script:
            pre_script = self.pre_script
        if not post_script:
            post_script = self.post_script
        if not python_cmd:
            python_cmd = self.python_cmd
        if not python_cmd:
            python_cmd = defaults.PYTHON_CMD
        if not pre_script and not post_script:
            return f"#!/usr/bin/env {python_cmd}\n{python_script}"
        if not pre_script:
            pre_script = ""
        if not post_script:
            post_script = ""
        return f"{pre_script}\ntype {python_cmd}\n{python_cmd} <<EOF\n{python_script}EOF\n\n{post_script}"



import os
from time import sleep

import pytest

from ..concurrent.futures import SlurmRestExecutor


@pytest.fixture(params=["tcp", "file"])
def slurm_executor_kwargs(
    request, slurm_data_directory, slurm_client_kwargs, slurm_parameters
):
    if request.param == "tcp":
        slurm_data_directory = None
    elif not slurm_parameters["mock"]:
        request.node.add_marker(
            pytest.mark.xfail(
                reason="File IO over distributed filesystems is not reliable"
            )
        )
    return {"data_directory": slurm_data_directory, **slurm_client_kwargs}


def test_rest_executor_submit(slurm_executor_kwargs):
    with SlurmRestExecutor(**slurm_executor_kwargs, max_workers=8) as slurm_executor:
        future1 = slurm_executor.submit(sleep, 0)
        future2 = slurm_executor.submit(sum, [1, 1])
        future3 = slurm_executor.submit(sum, [1, "a"])
        assert future1.result(timeout=60) is None
        assert future2.result(timeout=60) == 2
        with pytest.raises(TypeError):
            future3.result(timeout=60)


def test_rest_executor_map(slurm_executor_kwargs):
    with SlurmRestExecutor(**slurm_executor_kwargs, max_workers=8) as slurm_executor:
        results = [
            result
            for result in slurm_executor.map(sum, [[1, 1], [2, 2], [3, 3]], timeout=60)
        ]
    assert results == [2, 4, 6], str(results)


def test_rest_executor_initializer(slurm_executor_kwargs):
    with SlurmRestExecutor(
        **slurm_executor_kwargs,
        initializer=_initializer,
        max_workers=1,
        max_tasks_per_worker=5
    ) as slurm_executor:
        ftls = [slurm_executor.submit(_increment_global_value) for _ in range(5)]
        results = {future.result(timeout=60) for future in ftls}
    assert results == set(range(1, 6)), str(results)


def _initializer():
    global GLOBAL_VALUE
    GLOBAL_VALUE = 0


def _increment_global_value():
    global GLOBAL_VALUE
    GLOBAL_VALUE += 1
    return GLOBAL_VALUE


def test_rest_executor_jobs(slurm_executor_kwargs):
    with SlurmRestExecutor(
        **slurm_executor_kwargs, max_workers=8, max_tasks_per_worker=2
    ) as slurm_executor:
        ftls = [slurm_executor.submit(_job_ident) for _ in range(16)]
        results = {future.result(timeout=60) for future in ftls}
    assert len(results) == 8, str(len(results))


def _job_ident():
    return os.environ["SLURM_JOB_ID"]



import os
import time
import socket
import logging

from ._base import Client
from ._base import main_task_loop
from ._base import init_job


def main():
    init_job()
    with TcpClient() as client:
        main_task_loop(client)


class TcpClient(Client):
    def __init__(self) -> None:
        host = os.environ.get("_PYSLURMUTILS_HOST")
        port = int(os.environ.get("_PYSLURMUTILS_PORT"))
        try:
            hostname = socket.gethostbyaddr(host)[0]
        except Exception:
            hostname = host
        remote_name = f"{hostname}:{port}"

        self._client_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        logging.debug("Connecting to %s ...", remote_name)
        self._client_socket.settimeout(10)
        self._client_socket.connect((host, port))
        logging.debug("Connected to %s", remote_name)
        self._client_socket.settimeout(None)

        super().__init__(remote_name, remote_name)

    def close(self) -> None:
        self._client_socket.close()

    def _send_bytes(self, bdata):
        self._client_socket.sendall(bdata)

    def _receive_nbytes(self, nbytes):
        data = b""
        block = min(nbytes, 512)
        while len(data) < nbytes:
            data += self._client_socket.recv(block)
            time.sleep(0.1)
        return data



import os
import warnings
from typing import Optional

import pytest

from .mock_slurm import mock_slurm_clients


def pytest_addoption(parser):
    parser.addoption(
        "--slurm-root-directory",
        action="store",
        default=None,
        help="Specify the SLURM root directory for logs and data.",
    )


@pytest.fixture(scope="session")
def slurm_root_directory(tmp_path_factory, request):
    path = request.config.getoption("--slurm-root-directory")
    if path:
        return path
    return str(tmp_path_factory.mktemp("slurm_root"))


@pytest.fixture(scope="session")
def slurm_parameters(request) -> dict:
    url = os.environ.get("SLURM_URL", "mock")
    token = os.environ.get("SLURM_TOKEN", "mock")
    user_name = os.environ.get("SLURM_USER", "mock")
    mock = url == "mock" or token == "mock" or user_name == "mock"
    slurm_root_directory = request.config.getoption("--slurm-root-directory")
    if not mock and not slurm_root_directory:
        mock = True
        warnings.warn(
            "cannot use SLURM when the pytest argument '--slurm-root-directory ...' is not provided",
            UserWarning,
        )
    return {
        "url": url,
        "token": token,
        "user_name": user_name,
        "mock": mock,
    }


@pytest.fixture(scope="session")
def slurm_log_directory(
    slurm_root_directory, tmp_path_factory, slurm_parameters
) -> Optional[str]:
    if slurm_parameters["mock"]:
        return str(tmp_path_factory.mktemp("slurm_logs"))
    if slurm_root_directory:
        return os.path.join(
            slurm_root_directory, slurm_parameters["user_name"], "slurm_logs"
        )


@pytest.fixture(scope="session")
def slurm_data_directory(
    slurm_root_directory, tmp_path_factory, slurm_parameters
) -> Optional[str]:
    if slurm_parameters["mock"]:
        return str(tmp_path_factory.mktemp("slurm_data"))
    if slurm_root_directory:
        return os.path.join(
            slurm_root_directory, slurm_parameters["user_name"], "slurm_data"
        )


@pytest.fixture(scope="session")
def slurm_client_kwargs(
    slurm_log_directory, tmp_path_factory, slurm_parameters
) -> dict:
    params = dict(slurm_parameters)
    mock = params.pop("mock")
    params["log_directory"] = slurm_log_directory
    if mock:
        tmpdir = tmp_path_factory.mktemp("slurm_mock")
        with mock_slurm_clients(tmpdir):
            yield params
    else:
        yield params



import os
import time
import logging
import warnings
from uuid import uuid4

from ._connection_base import Connection
from ... import os_utils

logger = logging.getLogger(__name__)


class FileConnection(Connection):
    def __init__(self, directory: str, basename: str) -> None:
        conn_id = str(uuid4())
        os_utils.makedirs(directory)
        input_filename = os.path.join(directory, f"{basename}.in.{conn_id}.pkl")
        output_filename = os.path.join(directory, f"{basename}.out.{conn_id}.pkl")

        self._input_filename = input_filename
        self._output_filename = output_filename
        self._output_file = None
        self._input_file = open(input_filename, "wb+")

        logger.debug("start writing %s", input_filename)
        super().__init__()

    @property
    def input_name(self) -> str:
        return self._output_filename

    @property
    def output_name(self) -> str:
        return self._input_filename

    @property
    def input_filename(self) -> str:
        return self._input_filename

    @property
    def output_filename(self) -> str:
        return self._output_filename

    def close(self):
        if self._input_file is not None:
            self._input_file.close()
            self._input_file = None
            _delete_file_with_retry(self._input_filename)
        if self._output_file is not None:
            self._output_file.close()
            self._output_file = None
            _delete_file_with_retry(self._output_filename)

    def wait_client(self) -> None:
        if self._output_file is not None:
            return
        with self._wait_client_context() as it:
            dirname = os.path.dirname(self._output_filename)
            for _ in it:
                try:
                    _ = os.listdir(dirname)  # force NFS cache
                    self._output_file = open(self._output_filename, "rb+")
                    break
                except FileNotFoundError:
                    continue

    def receive_nbytes(self, nbytes: int) -> bytes:
        buffer = None
        with self._receive_nbytes_context(nbytes) as it:
            for buffer in it:
                if self._output_file is None:
                    self.wait_client()
                buffer.data += self._output_file.read(buffer.chunk_size)
        if buffer is None:
            raise RuntimeError(f"did not receive data from {self}")
        return buffer.data

    def send_bytes(self, data: bytes) -> None:
        if self.cancelled():
            raise RuntimeError(f"send to %{self} is cancelled")
        self._input_file.write(data)
        self._input_file.flush()


def _delete_file_with_retry(
    filename: str,
    max_attempts: int = 5,
    initial_delay: float = 0.1,
    max_delay: float = 1,
) -> None:
    delay = initial_delay / 2
    for _ in range(max_attempts):
        try:
            os.remove(filename)
            return
        except PermissionError:
            delay = min(delay * 2, max_delay)
            time.sleep(delay)
    warnings.warn(f"Could not remove file {filename}", UserWarning, stacklevel=2)



import random
import socket
import logging
from typing import Optional
from functools import lru_cache

from ._connection_base import Connection
from ... import defaults

logger = logging.getLogger(__name__)


class TcpConnection(Connection):
    def __init__(
        self,
        host: Optional[str] = None,
        port: Optional[int] = None,
        min_port: Optional[int] = None,
        num_ports: Optional[int] = None,
    ) -> None:
        if host is None:
            host = get_host()
        if port is None:
            self._server_sock = lock_port(host, min_port=min_port, num_ports=num_ports)
        else:
            self._server_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self._server_sock.bind((host, port))
        port = self._server_sock.getsockname()[-1]

        self._host = host
        self._port = port
        self._hostport = f"{host}:{port}"

        self._client_socket = None

        self._server_sock.listen(1)
        self._server_sock.settimeout(0.5)

        logger.debug("start listening on %s:%s", host, port)
        super().__init__()

    @property
    def input_name(self) -> str:
        return self._hostport

    @property
    def output_name(self) -> str:
        return self._hostport

    @property
    def host(self) -> str:
        return self._host

    @property
    def port(self) -> str:
        return self._port

    def close(self):
        if self._client_socket:
            self._client_socket.close()
        self._server_sock.close()

    def wait_client(self) -> None:
        if self._client_socket is not None:
            return
        with self._wait_client_context() as it:
            for _ in it:
                try:
                    self._client_socket, _ = self._server_sock.accept()
                    break
                except socket.timeout:
                    pass

    def receive_nbytes(self, nbytes: int) -> bytes:
        buffer = None
        with self._receive_nbytes_context(nbytes) as it:
            for buffer in it:
                if self._client_socket is None:
                    self.wait_client()
                buffer.data += self._client_socket.recv(buffer.chunk_size)
        if buffer is None:
            raise RuntimeError(f"did not receive data from {self.input_name}")
        return buffer.data

    def send_bytes(self, data: bytes) -> None:
        if self.cancelled():
            raise RuntimeError(f"send to {self.output_name} is cancelled")
        if self._client_socket is None:
            self.wait_client()
        self._client_socket.sendall(data)


@lru_cache(1)
def get_host() -> str:
    return socket.gethostbyname(socket.gethostname())


def lock_port(
    host: str, min_port: Optional[int] = None, num_ports: Optional[int] = None
) -> int:
    if min_port is None:
        min_port = defaults.MIN_PORT
    if num_ports is None:
        num_ports = defaults.NUM_PORTS
    preferred_ports = list(range(min_port, min_port + num_ports + 1))
    random.shuffle(preferred_ports)

    # Find a free port in the preferred range
    for port in preferred_ports:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            sock.bind((host, port))
            return sock
        except Exception:
            sock.close()
        except BaseException:
            sock.close()
            raise

    # Preferred ports are already in use, find any free port.
    while True:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            sock.bind((host, 0))
            return sock
        except Exception:
            sock.close()
        except BaseException:
            sock.close()
            raise
